#include "intersection.h"
#include <math.h>

using namespace dolt;

bool parallel(const Point& v1, const Point& v2);
bool onLine(const Point& p, const Point& vec, const Point& testp, float *t);
void lineIntersect(const Point& pA, const Point& vecA, 
		   const Point& pB, const Point& vecB, 
		   float *tA, float *tB);

const float closeEnough = 1e-5;

bool test_segmentIntersect(const Segment& a, const Segment& b)
{
  Point *p;
  Segment *s;
  segmentIntersect(a, b, &p, &s, false);
  if (p) {
    delete p;
    return true;
  }
  else if (s) {
    delete s;
    return true;
  }
  return false;
}


//
// line segment intersection (closed line segments)
//
// given two line segments a and b, find their intersection
// the intersection may be a point (pInt), a line segment (sInt)
// or the intersection may be null (pInt = sInt = 0)
//
// if there is an intersection, a Point or a Segment is created
// and the calling program must delete it when finished
//
// By default, closed line segments will be used.  This can be
// overridden with the optional parameter open
// 
void segmentIntersect(const Segment& a, const Segment& b,
		      Point **pInt, Segment **sInt,
		      bool open)
{ 
  Point vecA = a.end() - a.start();
  Point vecB = b.end() - b.start();
  float tA, tB;

  // default is "no intersection"
  *pInt = 0;
  *sInt = 0;
  
  if (parallel(vecA, vecB)) {
    // parallel line segments can intersect in a point or a line segment
    // check whether either endpoint of segment B is colinear with
    // segment A 
    float t1, t2;
    bool oneOn = onLine(a.start(), vecA, b.start(), &t1);
    bool twoOn = onLine(a.start(), vecA, b.end(), &t2);
    if (open) { 
      // if using open line segments, there will be an intersection
      // only if both points are colinear with segment A
      if (oneOn && twoOn) {
	float bMin, bMax;
	Point bMinP, bMaxP;
	// determine ordering w/r to vecA
	if (t1 < t2) {
	  bMin = t1; bMinP = b.start();
	  bMax = t2; bMaxP = b.end();
	}
	else {
	  bMin = t2; bMinP = b.end();
	  bMax = t1; bMaxP = b.start();
	}
	if (bMin < 1.0 && bMax > 0.0) {
	  // then we have a LineSegment intersection; determine the endpoints
	  Point left, right;
	  if (bMin > 0.0)
	    left = bMinP;
	  else
	    left = a.start();
	  if (bMax > 1.0)
	    right = a.end();
	  else
	    right = bMaxP;
	  *sInt = new Segment(left, right);
	  return;
	}
      }
      else 
	// segments are not colinear
	return;
    }
    else {
      // if using closed line segments
      if (!oneOn && !twoOn) 
	// neither endpoint is colinear with segment A
	return;
      else if (oneOn && !twoOn) {
	if (0.0 <= t1 && t1 <= 1.0) 
	  *pInt = new Point(b.start());
	return;
      }
      else if (!oneOn && twoOn) {
	if (0.0 <= t2 && t2 <= 1.0)
	  *pInt = new Point(b.end());
	return;
      }
      else { 
	// both b.start() and b.end() are colinear with segment A
	float bMin, bMax;
	Point bMinP, bMaxP;
	// determine the ordering w/r to vecA
	if (t1 < t2) {
	  bMin = t1; bMinP = b.start();
	  bMax = t2; bMaxP = b.end();
	}
	else {
	  bMin = t2; bMinP = b.end();
	  bMax = t1; bMaxP = b.start();
	}
	// no intersection cases
	if (bMin > 1 || bMax < 0)
	  return;
	// endpoint intersection cases
	else if (fabs(bMin-1) < closeEnough) {
	  *pInt = new Point(bMinP);
	  return;
	}
	else if (fabs(bMax) < closeEnough) {
	  *pInt = new Point(bMaxP);
	  return;
	}
	// segment intersection cases; take the max of the min, min of the max
	else {
	  Point left, right;
	  if (bMin > 0.0)
	    left = bMinP;
	  else
	    left = a.start();
	  if (bMax > 1.0)
	    right = a.end();
	  else
	    right = bMaxP;
	  *sInt = new Segment(left, right);
	  return;
	}
      }
    }
  }
  else {
    // lines are not perpendicular; intersect the two lines
    lineIntersect(a.start(), vecA, b.start(), vecB, &tA, &tB);
    if (open && (0.0 < tA) && (tA < 1.0) && (0.0 < tB) && (tB < 1.0)) {
      // the intersection point lies on the OPEN line segments
      // compute the point and return it
      *pInt = new Point(a.start() + (vecA * tA));
      return;
    }
    else if (!open && (0.0 <= tA) && (tA <= 1.0) && 
	     (0.0 <= tB) && (tB <= 1.0)) {
      // the intersection point lies on the CLOSED line segments
      // compute the point and return it
      *pInt = new Point(a.start() + (vecA * tA));
      return;
    }
    else {
      // the intersection point doesn't lie on either line segment
      return;
    }
  }
}


bool parallel(const Point& v1, const Point& v2)
{
  Point s1 = v1/v1.length();
  Point s2 = v2/v2.length();

  if (s1 == s2)
    return true;
  else if (s1*-1 == s2)
    return true;
  else
    return false;
}

//
// tests whether (and where) the point tp is on the line defined
// by the point p and vector vec
//
// returns true if so, and then t contains a value such tha p + vec*t == tp
//
// vec must not be the zero vector!
// 
bool onLine(const Point& p, const Point& vec, const Point& testp, float *t)
{
  // check if testp and p are the same point
  if (testp == p) {
    *t = 0.0;
    return true;
  }

  // otherwise, do the full check
  Point tvec = testp - p;

  if (parallel(tvec, vec)) {
    // "divide" tvec by vec but watch out for zero components of vec!
    if (fabs(vec.y()) < closeEnough)
      *t = tvec.x() / vec.x();
    else
      *t = tvec.y() / vec.y();
    return true;
  }
  else
    return false;
}

//
// find the intersection of two lines (it is assumed that they are not 
// parallel!)
//
// pA + tA * vecA == pB + tB * vecB
//
void lineIntersect(const Point& pA, const Point& vecA, 
		   const Point& pB, const Point& vecB, 
		   float *tA, float *tB)
{
  // using cramer's rule...
  float det = - vecA.x() * vecB.y() + vecA.y() * vecB.x();
  Point p = pB - pA;
  *tA = (- p.x() * vecB.y() + p.y() * vecB.x())/det;
  *tB = (vecA.x() * p.y() - vecA.y() * p.x())/det;
}
